@@ -12,6 +12,7 @@ from types import UnionType
 from typing import ClassVar
 from typing import Optional, Type, Any, Iterable, Callable, Sized
 
+import apispec.utils
 from apispec import APISpec
 from dict_caster.extras import first
 from http_tools import Answer
@@ -127,6 +128,8 @@ class GetIdsFromXlsxResponseSchema:
 
 
 class OpenApiWrapper:
+    openapi_version: ClassVar[str] = "3.1.0"
+
     @dataclasses.dataclass
     class Config:
         pass
@@ -142,8 +145,9 @@ class OpenApiWrapper:
         self.config = config
         self.context = context
         self.endpoints = []
+        self.required_dataclass_types: set[type[DataclassProtocol]] = set()
         self.dataclass_type_to_name: dict[type[DataclassProtocol] | UnionType, str] = {}
-        self.origin_type_to_factory: dict[type | UnionType, Callable[[APISpec, Iterable[type | UnionType]], dict]] = {
+        self.origin_type_to_factory: dict[type | UnionType, Callable[[Iterable[type | UnionType]], dict]] = {
             typing.Union: self._get_union_description,
             types.UnionType: self._get_union_description,
             list: self._get_list_description,
@@ -170,14 +174,22 @@ class OpenApiWrapper:
         spec = APISpec(
             title=f"{self.context.company}:{self.context.project_group}:{self.context.project_name}",
             version=self.version,
-            openapi_version="3.1.0"
+            openapi_version=self.openapi_version
         )
         for dataclass_type in self.dataclass_type_to_name:
-            self._add_schema_to_spec(spec, dataclass_type)
+            self._get_schema_description(dataclass_type)
         for endpoint in self.endpoints:
             self._add_endpoint_to_spec(spec, endpoint)
+        self._add_required_dataclass_types_to_spec(spec)
         return spec.to_dict()
 
+    def _add_required_dataclass_types_to_spec(self, spec: APISpec) -> None:
+        added_dataclass_types = set()
+        while missing_dataclass_types := self.required_dataclass_types - added_dataclass_types:
+            for dataclass_type in missing_dataclass_types:
+                self._add_dataclass_to_spec(spec, dataclass_type)
+                added_dataclass_types.add(dataclass_type)
+
     @staticmethod
     def _get_version() -> str:
         now = datetime.datetime.now()
@@ -188,18 +200,18 @@ class OpenApiWrapper:
             "summary": "",
             "operationId": endpoint.operation_id,
             "security": self._add_operation_securities(spec, endpoint.securities),
-            "parameters": self._add_operation_parameters(spec, endpoint.parameters),
+            "parameters": self._get_operation_parameters(endpoint.parameters),
             "responses": self._add_operation_code_to_answer(spec, endpoint.code_to_answer),
         }
-        if request_body_datum := self._add_operation_request_body(spec, endpoint.request_body):
+        if request_body_datum := self._get_operation_request_body(endpoint.request_body):
             operation_dict["requestBody"] = request_body_datum
 
         spec.path(endpoint.path, operations={endpoint.method: operation_dict})
 
-    def _add_operation_parameters(self, spec: APISpec, parameters: list[Parameter]) -> list[dict[str, Any]]:
+    def _get_operation_parameters(self, parameters: list[Parameter]) -> list[dict[str, Any]]:
         result: list[dict[str, Any]] = []
         for param in parameters:
-            schema_name = self._add_schema_to_spec(spec, param.schema, default=param.default)
+            schema_name = self._get_schema_description(param.schema, default=param.default)
             result.append({"name": param.name, "in": param.in_, "required": param.required, "schema": schema_name})
         return result
 
@@ -213,44 +225,44 @@ class OpenApiWrapper:
             result[str(int(code))] = response_name
         return result
 
-    def _add_operation_request_body(self, spec: APISpec, request_body: Optional[RequestBody]) -> dict[str, Any] | None:
+    def _get_operation_request_body(self, request_body: Optional[RequestBody]) -> dict[str, Any] | None:
         result = None
         if request_body is not None:
-            schema_name = self._add_schema_to_spec(spec, request_body.content.schema)
-            result = {"required": True, "content": {request_body.content.mime_type: {"schema": schema_name}}}
+            schema_description = self._get_schema_description(request_body.content.schema)
+            result = {"required": True, "content": {request_body.content.mime_type: {"schema": schema_description}}}
         return result
 
     @functools.cache
-    def _get_union_description(self, spec: APISpec, type_args: tuple[type | UnionType]) -> dict:
+    def _get_union_description(self, type_args: tuple[type | UnionType]) -> dict:
         assert type_args and len(type_args) >= 1, f'bad union element type: {type_args}'
-        return {"anyOf": [self._add_schema_to_spec(spec, arg) for arg in type_args]}
+        return {"anyOf": [self._get_schema_description(arg) for arg in type_args]}
 
     @functools.cache
-    def _get_list_description(self, spec: APISpec, type_args: tuple[type | UnionType] | None) -> dict:
+    def _get_list_description(self, type_args: tuple[type | UnionType] | None) -> dict:
         if type_args:
             assert type_args and len(type_args) == 1, f'bad list element type: {type_args}'
-            return {"type": "array", "items": self._add_schema_to_spec(spec, type_args[0])}
+            return {"type": "array", "items": self._get_schema_description(type_args[0])}
         else:
             return {"type": "array"}
 
-    def _get_dict_description(self, spec: APISpec, type_args: tuple[type | UnionType] | None) -> dict:
+    def _get_dict_description(self, type_args: tuple[type | UnionType] | None) -> dict:
         if type_args:
             assert len(type_args) == 2, f'bad dict item type: {type_args}'
-            return {"type": "object", "additionalProperties": self._add_schema_to_spec(spec, type_args[1])}
+            return {"type": "object", "additionalProperties": self._get_schema_description(type_args[1])}
         else:
             return {"type": "object"}
 
     def _get_dataclass_description(self, spec: APISpec, schema: DataclassProtocol) -> dict:
         return {"type": "object", "properties": {
-            field_.name: self._add_schema_to_spec(spec, field_.type)
+            field_.name: self._get_schema_description(field_.type)
             for field_ in dataclasses.fields(schema) if field_.repr
         }}
 
     @functools.cache
     def _get_generic_description(
-            self, spec: APISpec, origin_type: type | UnionType, type_args: tuple[type | UnionType] | None) -> dict:
+            self, origin_type: type | UnionType, type_args: tuple[type | UnionType] | None) -> dict:
         if factory := self.origin_type_to_factory.get(origin_type):
-            return factory(spec, type_args)
+            return factory(type_args)
         else:
             raise TypeError(f"Unknown origin_type: {origin_type}")
 
@@ -268,33 +280,34 @@ class OpenApiWrapper:
         raise TypeError(f"Unknown type: {type_}")
 
     @functools.cache
-    def _add_schema_to_spec(
-            self, spec: APISpec, schema: type | UnionType, default: Any = NoValue, base: dict | None = None
-    ) -> dict:
-        base = base or {}
-        logger.debug('_add_schema_to_spec: %s', schema)
-
-        origin_type = typing.get_origin(schema)
-        if origin_type is not None:
-            description = self._get_generic_description(spec, origin_type, typing.get_args(schema))
+    def _get_schema_description(self, schema: type | UnionType, default: Any = NoValue) -> dict:
+        logger.debug('_get_schema_description: %s', schema)
+        if origin_type := typing.get_origin(schema):
+            description = self._get_generic_description(origin_type, typing.get_args(schema))
         elif schema in self.origin_type_to_factory:
-            description = self._get_generic_description(spec, schema, None)
+            description = self._get_generic_description(schema, None)
         elif dataclasses.is_dataclass(schema):
-            description = self._get_dataclass_description(spec, schema)
+            self.required_dataclass_types.add(schema)
             name = self.dataclass_type_to_name.get(schema, f"schema_{schema.__name__}")
-            return self._save_schema_to_spec(spec, description, name)
+            return apispec.utils.build_reference("schema", 3, name)
         else:
             description = self._get_primitive_description(schema)
 
         if isinstance(schema, type) and issubclass(schema, Enum):
             description['enum'] = [e.value for e in schema]
-
         if default is None or isinstance(default, (int, float, str, list, dict)):
             description['default'] = default
         # if default == NoValue:  # TODO: think about it, should we place "required" inside schema properties or not?
         #     description['required'] = True
+        return description
 
-        return base | description
+    @functools.cache
+    def _add_dataclass_to_spec(self, spec: APISpec, schema: DataclassProtocol) -> None:
+        logger.debug(f'_add_dataclass_to_spec: {schema}')
+        assert dataclasses.is_dataclass(schema)
+        description = self._get_dataclass_description(spec, schema)
+        name = self.dataclass_type_to_name.get(schema, f"schema_{schema.__name__}")
+        spec.components.schema(component_id=name, component=description)
 
     @functools.cache
     def _add_answer_to_spec(self, spec: APISpec, answer: Type[Answer]) -> str:
@@ -304,7 +317,7 @@ class OpenApiWrapper:
         # response_schema = response.__args__[0]
         # response_type = response.__origin__
         # name = f"response_{response_type.__name__}[{response_schema.__name__}]"
-        schema_name_or_description = self._add_schema_to_spec(spec, schema)
+        schema_name_or_description = self._get_schema_description(schema)
 
         spec.components.response(
             component_id=name, component={
